from __future__ import annotations

import pytest

import flexparser.flexparser as fp


@pytest.mark.parametrize(
    "delimiters,content,expected",
    [
        # ### 0
        (
            {},
            "Testing # 123",
            ((0, 0, 0, 13, "Testing # 123"),),
        ),
        # ### 1
        (
            {"#": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.STOP_PARSING)},
            "Testing # 123",
            ((0, 0, 0, 8, "Testing "),),
        ),
        # ### 2
        (
            {"#": (fp.DelimiterInclude.SPLIT_AFTER, fp.DelimiterAction.STOP_PARSING)},
            "Testing # 123",
            ((0, 0, 0, 9, "Testing #"),),
        ),
        # ### 3
        (
            {"#": (fp.DelimiterInclude.SPLIT_BEFORE, fp.DelimiterAction.STOP_PARSING)},
            "Testing # 123",
            ((0, 0, 0, 8, "Testing "),),
        ),
    ],
)
def test_split_single_line(delimiters, content, expected):
    out = tuple(fp.Spliter(content, delimiters))
    assert out == expected


@pytest.mark.parametrize(
    "delimiters,content,expected",
    [
        # ### 0
        (
            {},
            "Testing # 123\nCaption # 456",
            ((0, 0, 1, 13, "Testing # 123\nCaption # 456"),),
        ),
        # ### 1
        (
            {"#": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.STOP_PARSING)},
            "Testing # 123\nCaption # 456",
            ((0, 0, 0, 8, "Testing "),),
        ),
        # ### 2
        (
            {"#": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.STOP_PARSING_LINE)},
            "Testing # 123\nCaption # 456",
            (
                (0, 0, 0, 8, "Testing "),
                (1, 0, 1, 8, "Caption "),
            ),
        ),
        # ### 3
        (
            {"#": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL)},
            "Testing # 123\nCaption # 456",
            (
                (0, 0, 0, 8, "Testing "),
                (0, 9, 1, 8, " 123\nCaption "),
                (1, 9, 1, 13, " 456"),
            ),
        ),
        # ### 4
        (
            {
                "#": (
                    fp.DelimiterInclude.SPLIT_BEFORE,
                    fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
                ),
                "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
            },
            "Testing # 123\nCaption # 456",
            (
                (0, 0, 0, 8, "Testing "),
                (0, 8, 0, 13, "# 123"),
                (1, 0, 1, 8, "Caption "),
                (1, 8, 1, 13, "# 456"),
            ),
        ),
        # ### 5
        (
            {
                "#": (
                    fp.DelimiterInclude.SPLIT_BEFORE,
                    fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
                ),
                "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
            },
            "Testing ## 123\nCaption ## 456",
            (
                (0, 0, 0, 8, "Testing "),
                (0, 8, 0, 14, "## 123"),
                (1, 0, 1, 8, "Caption "),
                (1, 8, 1, 14, "## 456"),
            ),
        ),
        # ### 6
        (
            {
                "#": (fp.DelimiterInclude.SPLIT_BEFORE, fp.DelimiterAction.CONTINUE),
                "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
            },
            "Testing ## 123\nCaption ## 456",
            (
                (0, 0, 0, 8, "Testing "),
                (0, 8, 0, 9, "#"),
                (0, 9, 0, 14, "# 123"),
                (1, 0, 1, 8, "Caption "),
                (1, 8, 1, 9, "#"),
                (1, 9, 1, 14, "# 456"),
            ),
        ),
        # ### 7
        (
            {
                "#": (
                    fp.DelimiterInclude.SPLIT_BEFORE,
                    fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
                ),
                "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
                "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
            },
            "Testing ## 123\nCaption ## 456",
            (
                (0, 0, 0, 8, "Testing "),
                (0, 8, 0, 14, "## 123"),
                (1, 0, 1, 8, "Caption "),
                (1, 8, 1, 14, "## 456"),
            ),
        ),
    ],
)
def test_split_multi_line(delimiters, content, expected):
    out = tuple(fp.Spliter(content, delimiters))
    assert out == expected


def test_statement():
    dlm = {
        "#": (
            fp.DelimiterInclude.SPLIT_BEFORE,
            fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
        ),
        "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
    }
    content = "Testing ## 123\nCaption ## 456"
    bi = fp.StatementIterator(content, dlm)
    assert bi.peek().raw_strip == "Testing"
    assert next(bi).raw_strip == "Testing"
    assert bi.peek().raw_strip == "## 123"
    assert next(bi).raw_strip == "## 123"

    el = next(bi)
    # strip spaces now changes the element
    # not the parser.
    assert el.raw == "Caption"
    assert el.raw_strip == "Caption"
    assert el.start_line == 2
    assert el.start_col == 0
    assert el.end_line == 2
    assert el.end_col == 7

    assert next(bi).raw_strip == "## 456"
    assert bi.peek("blip") == "blip"
    with pytest.raises(StopIteration):
        bi.peek()
    with pytest.raises(StopIteration):
        next(bi)


def test_statement2():
    dlm = {
        "#": (
            fp.DelimiterInclude.SPLIT_BEFORE,
            fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
        ),
        "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
    }
    content = "Testing ## 123\nCaption ## 456"
    bi = fp.StatementIterator(content, dlm)
    assert bi.peek().raw_strip == "Testing"
    assert next(bi).raw_strip == "Testing"
    assert bi.peek().raw_strip == "## 123"
    assert next(bi).raw_strip == "## 123"

    el = next(bi)
    # strip spaces now changes the element
    # not the parser.
    assert el.raw == "Caption"
    assert el.raw_strip == "Caption"
    assert el.start_line == 2
    assert el.start_col == 0
    assert el.end_line == 2
    assert el.end_col == 7

    assert next(bi).raw_strip == "## 456"
    assert bi.peek("blip") == "blip"
    with pytest.raises(StopIteration):
        bi.peek()
    with pytest.raises(StopIteration):
        next(bi)


def test_statement_change_dlm():
    dlm = {
        "#": (
            fp.DelimiterInclude.SPLIT_BEFORE,
            fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
        ),
        "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
    }

    dlm_new = {
        "!": (
            fp.DelimiterInclude.SPLIT_BEFORE,
            fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
        ),
        "\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
        "\r\n": (fp.DelimiterInclude.SPLIT, fp.DelimiterAction.CONTINUE),
    }
    content = "Testing ## 123\nCaption !! 456"
    bi = fp.StatementIterator(content, dlm)
    assert bi.peek().raw_strip == "Testing"
    assert next(bi).raw_strip == "Testing"
    assert bi.peek().raw_strip == "## 123"
    assert next(bi).raw_strip == "## 123"

    assert bi.peek().raw_strip == "Caption !! 456"
    bi.set_delimiters(dlm_new)
    assert bi.peek().raw_strip == "Caption"
    assert next(bi).raw_strip == "Caption"
    assert next(bi).raw_strip == "!! 456"



"""
flexparser
~~~~~~~~~

Classes and functions to create parsers.

The idea is quite simple. You write a class for every type of content
(called here ``ParsedStatement``) you need to parse. Each class should
have a ``from_string`` constructor. We used extensively the ``typing``
module to make the output structure easy to use and less error prone.

:copyright: 2022 by flexparser Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from importlib.metadata import version

try:  # pragma: no cover
    __version__ = version("flexparser")
except Exception:  # pragma: no cover
    # we seem to have a local copy not installed without setuptools
    # so the reported version will be unknown
    __version__ = "unknown"


from .flexparser import (
    BOF,
    BOR,
    BOS,
    EOS,
    SPLIT_EOL,
    Block,
    DelimiterAction,
    DelimiterInclude,
    IncludeStatement,
    NullableParsedResult,
    ParsedProject,
    ParsedSource,
    ParsedStatement,
    Parser,
    ParsingError,
    RootBlock,
    StatementIterator,
    UnexpectedEOS,
    UnknownStatement,
    parse,
    parse_bytes,
)

# Deprecate in 0.3
UnexpectedEOF = UnexpectedEOS

__all__ = (
    "__version__",
    "Block",
    "DelimiterAction",
    "DelimiterInclude",
    "IncludeStatement",
    "ParsedStatement",
    "Parser",
    "ParsingError",
    "RootBlock",
    "StatementIterator",
    "UnexpectedEOF",
    "UnexpectedEOS",
    "UnknownStatement",
    "parse",
    "ParsedSource",
    "ParsedProject",
    "SPLIT_EOL",
    "BOF",
    "BOR",
    "BOS",
    "EOS",
    "parse_bytes",
    "NullableParsedResult",
)



from __future__ import annotations

import hashlib
import pathlib
from dataclasses import dataclass

import pytest

from flexparser import flexparser as fp
from flexparser.testsuite.common import (
    Close,
    Comment,
    EqualFloat,
    MyBlock,
    MyParser,
    MyRoot,
    Open,
)

FIRST_NUMBER = 1


def _bosser(content: bytes):
    return fp.BOS(fp.Hash.from_bytes(hashlib.blake2b, content))


def _compare(arr1, arr2):
    assert len(arr1) == len(arr2)
    for a1, a2 in zip(arr1, arr2):
        if isinstance(a1, fp.BOS) and isinstance(a2, fp.BOS):
            assert a1.content_hash == a2.content_hash
        else:
            assert a1 == a2, str(a1) + " == " + str(a2)


def test_locator():
    this_file = pathlib.Path(__file__)

    with pytest.raises(ValueError):
        # Cannot use absolute path as target.
        assert fp.default_locator(this_file, "/temp/bla.txt")

    with pytest.raises(TypeError):
        assert fp.default_locator(str(this_file), "bla.txt")

    with pytest.raises(TypeError):
        assert fp.default_locator(str(this_file), "/temp/bla.txt")

    assert fp.default_locator(this_file, "bla.txt") == this_file.parent / "bla.txt"
    assert (
        fp.default_locator(this_file.parent, "bla.txt") == this_file.parent / "bla.txt"
    )

    with pytest.raises(ValueError):
        assert (
            fp.default_locator(this_file.parent, "../bla.txt")
            == this_file.parent / "bla.txt"
        )

    assert fp.default_locator(("pack", "nam"), "bla") == ("pack", "bla")


@pytest.mark.parametrize("definition", [MyRoot, (Comment, EqualFloat), MyParser])
def test_parse1(tmp_path, definition):
    content = b"# hola\nx=1.0"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_bytes(content)

    pp = fp.parse(tmp_file, definition)

    assert len(pp) == 1

    psf = pp[list(pp.keys())[0]]
    assert not psf.has_errors
    assert psf.config is None
    assert psf.parsed_source.opening.mtime == tmp_file.stat().st_mtime
    assert psf.parsed_source.opening.path == tmp_file
    assert tuple(psf.errors()) == ()
    assert psf.location == tmp_file

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 1, 0, 5)
        .set_raw("x=1.0"),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors

    _compare(
        tuple(pp.iter_statements()),
        (
            _bosser(content).set_simple_position(FIRST_NUMBER + 0, 0, 0),
            Comment("# hola")
            .set_simple_position(FIRST_NUMBER + 0, 0, 6)
            .set_raw("# hola"),
            EqualFloat("x", 1.0)
            .set_simple_position(FIRST_NUMBER + 1, 0, 5)
            .set_raw("x=1.0"),
            fp.EOS().set_simple_position(FIRST_NUMBER + 2, 0, 0),
        ),
    )


@pytest.mark.parametrize("definition", [MyRoot, EqualFloat, MyParser])
def test_parse2(tmp_path, definition):
    content = b"y = 2.0\nx=1.0"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_bytes(content)

    pp = fp.parse(tmp_file, definition)

    assert len(pp) == 1

    assert None in pp
    psf = pp[None]
    assert not psf.has_errors
    assert psf.config is None
    assert psf.parsed_source.opening.mtime == tmp_file.stat().st_mtime
    assert psf.parsed_source.opening.path == tmp_file
    assert tuple(psf.errors()) == ()
    assert psf.location == tmp_file

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        EqualFloat("y", 2.0)
        .set_simple_position(FIRST_NUMBER + 0, 0, 7)
        .set_raw("y = 2.0"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 1, 0, 5)
        .set_raw("x=1.0"),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors

    _compare(
        tuple(pp.iter_statements()),
        (
            _bosser(content).set_simple_position(FIRST_NUMBER + 0, 0, 0),
            EqualFloat("y", 2.0)
            .set_simple_position(FIRST_NUMBER + 0, 0, 7)
            .set_raw("y = 2.0"),
            EqualFloat("x", 1.0)
            .set_simple_position(FIRST_NUMBER + 1, 0, 5)
            .set_raw("x=1.0"),
            fp.EOS().set_simple_position(FIRST_NUMBER + 2, 0, 0),
        ),
    )


@pytest.mark.parametrize(
    "definition",
    [
        MyBlock,
    ],
)
def test_parse3(tmp_path, definition):
    content = b"@begin\ny = 2.0\nx=1.0\n@end"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_bytes(content)

    pp = fp.parse(tmp_file, definition)
    assert not pp.has_errors
    assert len(pp) == 1
    assert tuple(pp.errors()) == ()

    assert None in pp
    psf = pp[None]
    assert not psf.has_errors
    assert psf.config is None
    assert psf.parsed_source.opening.mtime == tmp_file.stat().st_mtime
    assert psf.parsed_source.opening.path == tmp_file
    assert tuple(psf.errors()) == ()
    assert psf.location == tmp_file

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 1
    mb = body[0]

    assert mb.start_line == 1
    assert mb.start_col == 0
    assert mb.end_line == 4
    assert mb.end_col == 4
    assert mb.format_position == "1,0-4,4"

    assert mb.opening == Open().set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw(
        "@begin"
    )
    assert tuple(mb.body) == (
        EqualFloat("y", 2.0)
        .set_simple_position(FIRST_NUMBER + 1, 0, 7)
        .set_raw("y = 2.0"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 2, 0, 5)
        .set_raw("x=1.0"),
    )
    assert mb.closing == Close().set_simple_position(FIRST_NUMBER + 3, 0, 4).set_raw(
        "@end"
    )
    assert not mb.has_errors

    _compare(
        tuple(pp.iter_statements()),
        (
            _bosser(content).set_simple_position(FIRST_NUMBER + 0, 0, 0),
            Open().set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw("@begin"),
            EqualFloat("y", 2.0)
            .set_simple_position(FIRST_NUMBER + 1, 0, 7)
            .set_raw("y = 2.0"),
            EqualFloat("x", 1.0)
            .set_simple_position(FIRST_NUMBER + 2, 0, 5)
            .set_raw("x=1.0"),
            Close().set_simple_position(FIRST_NUMBER + 3, 0, 4).set_raw("@end"),
            fp.EOS().set_simple_position(FIRST_NUMBER + 4, 0, 0),
        ),
    )


def test_include_file(tmp_path):
    @dataclass(frozen=True)
    class Include(fp.IncludeStatement[None]):
        value: str

        @property
        def target(self) -> str:
            return "bla2.txt"

        @classmethod
        def from_string(cls, s: str):
            if s.startswith("include"):
                return cls(s[len("include ") :].strip())

    content1 = b"include bla2.txt\n# chau"
    content2 = b"# hola\nx=1.0"

    tmp_file1 = tmp_path / "bla1.txt"
    tmp_file2 = tmp_path / "bla2.txt"
    tmp_file1.write_bytes(content1)
    tmp_file2.write_bytes(content2)

    pp = fp.parse(tmp_file1, (Include, Comment, EqualFloat), None)

    assert None in pp
    assert (tmp_file1, "bla2.txt") in pp

    assert len(pp) == 2

    _compare(
        tuple(pp.iter_statements()),
        (
            _bosser(content1).set_simple_position(FIRST_NUMBER + 0, 0, 0),
            # Include
            _bosser(content2).set_simple_position(FIRST_NUMBER + 0, 0, 6),
            Comment("# hola")
            .set_simple_position(FIRST_NUMBER + 0, 0, 6)
            .set_raw("# hola"),
            EqualFloat("x", 1.0)
            .set_simple_position(FIRST_NUMBER + 1, 0, 5)
            .set_raw("x=1.0"),
            fp.EOS().set_simple_position(FIRST_NUMBER + 2, 0, 0),
            Comment("# chau")
            .set_simple_position(FIRST_NUMBER + 1, 0, 6)
            .set_raw("# chau"),
            fp.EOS().set_simple_position(FIRST_NUMBER + 2, 0, 0),
        ),
    )


def test_resources(tmp_path):
    @dataclass(frozen=True)
    class Include(fp.IncludeStatement[None]):
        value: str

        @property
        def target(self) -> str:
            return "bla2.txt"

        @classmethod
        def from_string(cls, s: str):
            if s.startswith("include"):
                return cls(s[len("include ") :].strip())

    # see files included in the textsuite.
    content1 = b"include bla2.txt\n# chau\n"
    content2 = b"# hola\nx=1.0\n"

    pp = fp.parse(("flexparser.testsuite", "bla1.txt"), (Include, Comment, EqualFloat))

    assert len(pp) == 2

    _compare(
        tuple(pp.iter_statements()),
        (
            _bosser(content1).set_simple_position(FIRST_NUMBER + 0, 0, 0),
            # include
            _bosser(content2).set_simple_position(FIRST_NUMBER + 0, 0, 0),
            Comment("# hola")
            .set_simple_position(FIRST_NUMBER + 0, 0, 6)
            .set_raw("# hola"),
            EqualFloat("x", 1.0)
            .set_simple_position(FIRST_NUMBER + 1, 0, 5)
            .set_raw("x=1.0"),
            fp.EOS().set_simple_position(FIRST_NUMBER + 2, 0, 0),
            Comment("# chau")
            .set_simple_position(FIRST_NUMBER + 1, 0, 6)
            .set_raw("# chau"),
            fp.EOS().set_simple_position(FIRST_NUMBER + 2, 0, 0),
        ),
    )



from __future__ import annotations

import pytest

from flexparser import flexparser as fp
from flexparser.testsuite.common import (
    Comment,
    EqualFloat,
    MyBlock,
    MyParser,
    MyParserWithBlock,
)

FIRST_NUMBER = 1


@pytest.mark.parametrize(
    "content", (b"# hola\nx<>1.0", b"# hola\r\nx<>1.0", b"# hola\rx<>1.0")
)
def test_consume_err(content):
    myparser = MyParser(None)

    pf = myparser.parse_bytes(content).parsed_source
    assert isinstance(pf.opening, fp.BOS)
    assert isinstance(pf.closing, fp.EOS)
    body = tuple(pf.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw("# hola"),
        fp.UnknownStatement()
        .set_simple_position(FIRST_NUMBER + 1, 0, 6)
        .set_raw("x<>1.0"),
    )
    assert tuple(pf) == (pf.opening, *body, pf.closing)
    assert pf.has_errors

    assert str(body[-1]) == "Could not parse 'x<>1.0' (2,0-2,6)"


@pytest.mark.parametrize(
    "content", (b"# hola\nx=1.0", b"# hola\r\nx=1.0", b"# hola\rx=1.0")
)
def test_consume(content):
    myparser = MyParser(None)

    pf = myparser.parse_bytes(content).parsed_source
    assert pf.start_line == 0
    assert pf.start_col == 0
    assert pf.end_line == 3
    assert pf.end_col == 0
    assert pf.format_position == "0,0-3,0"
    assert isinstance(pf.opening, fp.BOS)
    assert isinstance(pf.closing, fp.EOS)
    body = tuple(pf.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 1, 0, 5)
        .set_raw("x=1.0"),
    )
    assert tuple(pf) == (pf.opening, *body, pf.closing)
    assert not pf.has_errors


@pytest.mark.parametrize("use_string", (True, False))
def test_parse(tmp_path, use_string):
    content = "# hola\nx=1.0"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_text(content)
    myparser = MyParser(None)

    if use_string:
        psf = myparser.parse(str(tmp_file))
    else:
        psf = myparser.parse(tmp_file)

    assert not psf.has_errors
    assert psf.config is None
    assert psf.parsed_source.opening.mtime == tmp_file.stat().st_mtime
    assert psf.parsed_source.opening.path == tmp_file
    assert tuple(psf.errors()) == ()
    assert psf.location == tmp_file

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 1, 0, 5)
        .set_raw("x=1.0"),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors


def test_unfinished_block(tmp_path):
    content = "@begin\n# hola\nx=1.0"

    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_text(content)
    myparser = MyParserWithBlock(None)

    psf = myparser.parse(tmp_file)
    assert psf.has_errors
    assert psf.config is None
    assert psf.parsed_source.opening.mtime == tmp_file.stat().st_mtime
    assert psf.parsed_source.opening.path == tmp_file
    assert tuple(psf.errors()) == (
        fp.UnexpectedEOS().set_simple_position(FIRST_NUMBER + 3, 0, 0),
    )
    assert psf.location == tmp_file

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 1
    assert isinstance(body[0], MyBlock)
    assert body[0].closing == fp.UnexpectedEOS().set_simple_position(
        FIRST_NUMBER + 3, 0, 0
    )
    assert body[0].body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 1, 0, 6).set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 2, 0, 5)
        .set_raw("x=1.0"),
    )


@pytest.mark.parametrize("use_resource", (True, False))
def test_parse_resource(use_resource):
    myparser = MyParser(None)

    location = ("flexparser.testsuite", "bla2.txt")

    if use_resource:
        psf = myparser.parse_resource(*location)
    else:
        psf = myparser.parse(location)

    assert not psf.has_errors
    assert psf.config is None
    assert tuple(psf.errors()) == ()

    if use_resource:
        assert psf.location == location

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 1, 0, 5)
        .set_raw("x=1.0"),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Union

if sys.version_info >= (3, 10):
    from typing import TypeAlias  # noqa
else:
    from typing_extensions import TypeAlias  # noqa


if sys.version_info >= (3, 11):
    from typing import Self  # noqa
else:
    from typing_extensions import Self  # noqa

from flexparser import flexparser as fp


class NotAValidIdentifier(fp.ParsingError):
    value: str

    def __init__(self, value: str) -> None:
        self.value = value

    def custom_values_str(self) -> str:
        return f"value='{self.value}'"


class CannotParseToFloat(fp.ParsingError):
    value: str

    def __init__(self, value: str) -> None:
        self.value = value

    def custom_values_str(self) -> str:
        return f"value='{self.value}'"


@dataclass(frozen=True)
class Open(fp.ParsedStatement[None]):
    @classmethod
    def from_string(cls, s: str) -> fp.NullableParsedResult[Self]:
        if s == "@begin":
            return cls()
        return None


@dataclass(frozen=True)
class Close(fp.ParsedStatement[None]):
    @classmethod
    def from_string(cls, s: str) -> fp.NullableParsedResult[Self]:
        if s == "@end":
            return cls()
        return None


@dataclass(frozen=True)
class Comment(fp.ParsedStatement[None]):
    s: str

    @classmethod
    def from_string(cls, s: str) -> fp.NullableParsedResult[Self]:
        if s.startswith("#"):
            return cls(s)
        return None


@dataclass(frozen=True)
class EqualFloat(fp.ParsedStatement[None]):
    a: str
    b: float

    @classmethod
    def from_string(cls, s: str) -> fp.NullableParsedResult[Self]:
        if "=" not in s:
            return None

        a, b = s.split("=")
        a = a.strip()
        b = b.strip()

        if not str.isidentifier(a):
            return NotAValidIdentifier(a)

        try:
            b = float(b)
        except Exception:
            return CannotParseToFloat(b)

        return cls(a, b)


class MyBlock(fp.Block[Open, Union[Comment, EqualFloat], Close, None]):
    pass


class MyRoot(fp.RootBlock[Union[Comment, EqualFloat], None]):
    pass


class MyParser(fp.Parser[MyRoot, None]):
    pass


class MyRootWithBlock(fp.RootBlock[Union[Comment, EqualFloat, MyBlock], None]):
    pass


class MyParserWithBlock(fp.Parser[MyRootWithBlock, None]):
    pass



from __future__ import annotations

from typing import Union

import pytest

from flexparser import flexparser as fp
from flexparser.testsuite.common import (
    CannotParseToFloat,
    Close,
    Comment,
    EqualFloat,
    MyBlock,
    MyRoot,
    NotAValidIdentifier,
    Open,
)


class MyBlock2(fp.Block[Open, Union[Comment, EqualFloat], Close, None]):
    pass


class MyRoot2(fp.RootBlock[Union[Comment, EqualFloat], None]):
    pass


FIRST_NUMBER = 1


def test_block_classes():
    assert tuple(MyBlock.opening_classes()) == (Open,)
    assert tuple(MyBlock.body_classes()) == (
        Comment,
        EqualFloat,
    )
    assert tuple(MyBlock.closing_classes()) == (Close,)

    assert tuple(MyRoot.opening_classes()) == ()
    assert tuple(MyRoot.body_classes()) == (
        Comment,
        EqualFloat,
    )
    assert tuple(MyRoot.closing_classes()) == ()

    assert tuple(MyBlock2.opening_classes()) == (Open,)
    assert tuple(MyBlock2.body_classes()) == (
        Comment,
        EqualFloat,
    )
    assert tuple(MyBlock2.closing_classes()) == (Close,)

    assert tuple(MyRoot2.opening_classes()) == ()
    assert tuple(MyRoot2.body_classes()) == (
        Comment,
        EqualFloat,
    )
    assert tuple(MyRoot2.closing_classes()) == ()


def test_formatting():
    obj = EqualFloat.from_string("a = 3.1")
    assert obj is not None
    assert obj.format_position == "N/A"
    obj.set_simple_position(10, 3, 7)
    assert obj.format_position == "10,3-10,10"
    assert (
        str(obj)
        == "EqualFloat(start_line=10, start_col=3, end_line=10, end_col=10, raw=None, a='a', b=3.1)"
    )

    obj = EqualFloat.from_string("%a = 3.1")
    assert obj is not None
    assert obj.format_position == "N/A"
    obj.set_simple_position(10, 3, 8)
    assert obj.format_position == "10,3-10,11"


def test_parse_equal_float():
    assert EqualFloat.from_string("a = 3.1") == EqualFloat("a", 3.1)
    assert EqualFloat.from_string("a") is None

    assert EqualFloat.from_string("%a = 3.1") == NotAValidIdentifier("%a")
    assert EqualFloat.from_string("a = 3f1") == CannotParseToFloat("3f1")

    obj = EqualFloat.from_string("a = 3f1")
    assert (
        str(obj)
        == "CannotParseToFloat(start_line=0, start_col=0, end_line=0, end_col=0, raw=None, value='3f1')"
    )


def test_consume_equal_float():
    f = lambda s: fp.StatementIterator(s, fp.SPLIT_EOL)
    assert EqualFloat.consume(f("a = 3.1"), None) == EqualFloat("a", 3.1).set_position(
        1, 0, 1, 7
    ).set_raw("a = 3.1")
    assert EqualFloat.consume(f("a"), None) is None

    assert EqualFloat.consume(f("%a = 3.1"), None) == NotAValidIdentifier(
        "%a"
    ).set_position(1, 0, 1, 8).set_raw("%a = 3.1")
    assert EqualFloat.consume(f("a = 3f1"), None) == CannotParseToFloat(
        "3f1"
    ).set_position(1, 0, 1, 7).set_raw("a = 3f1")


@pytest.mark.parametrize("klass", (MyRoot, MyRoot2))
def test_stream_block(klass):
    lines = "# hola\nx=1.0"
    si = fp.StatementIterator(lines, fp.SPLIT_EOL)

    mb = klass.consume_body_closing(fp.BOS(fp.Hash.nullhash()), si, None)
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola")
        .set_position(FIRST_NUMBER + 0, 0, FIRST_NUMBER + 0, 6)
        .set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_position(FIRST_NUMBER + 1, 0, FIRST_NUMBER + 1, 5)
        .set_raw("x=1.0"),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors


@pytest.mark.parametrize("klass", (MyRoot, MyRoot2))
def test_stream_block_error(klass):
    lines = "# hola\nx=1f0"
    si = fp.StatementIterator(lines, fp.SPLIT_EOL)

    mb = klass.consume_body_closing(fp.BOS(fp.Hash.nullhash()), si, None)
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw("# hola"),
        CannotParseToFloat("1f0")
        .set_simple_position(FIRST_NUMBER + 1, 0, 5)
        .set_raw("x=1f0"),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert mb.has_errors
    assert mb.errors == (
        CannotParseToFloat("1f0")
        .set_simple_position(FIRST_NUMBER + 1, 0, 5)
        .set_raw("x=1f0"),
    )


@pytest.mark.parametrize("klass", (MyBlock, MyBlock2))
def test_block(klass):
    lines = "@begin\n# hola\nx=1.0\n@end"
    si = fp.StatementIterator(lines, fp.SPLIT_EOL)

    mb = klass.consume(si, None)
    assert mb.opening == Open().set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw(
        "@begin"
    )
    assert mb.closing == Close().set_simple_position(FIRST_NUMBER + 3, 0, 4).set_raw(
        "@end"
    )
    body = tuple(mb.body)
    assert len(body) == 2
    assert mb.body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 1, 0, 6).set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 2, 0, 5)
        .set_raw("x=1.0"),
    )

    assert tuple(mb) == (mb.opening, *mb.body, mb.closing)
    assert not mb.has_errors


@pytest.mark.parametrize("klass", (MyBlock, MyBlock2))
def test_unfinished_block(klass):
    lines = "@begin\n# hola\nx=1.0"
    si = fp.StatementIterator(lines, fp.SPLIT_EOL)

    mb = klass.consume(si, None)
    assert mb.opening == Open().set_simple_position(FIRST_NUMBER + 0, 0, 6).set_raw(
        "@begin"
    )
    assert mb.closing == fp.UnexpectedEOS().set_simple_position(FIRST_NUMBER + 3, 0, 0)
    body = tuple(mb.body)
    assert len(body) == 2
    assert mb.body == (
        Comment("# hola").set_simple_position(FIRST_NUMBER + 1, 0, 6).set_raw("# hola"),
        EqualFloat("x", 1.0)
        .set_simple_position(FIRST_NUMBER + 2, 0, 5)
        .set_raw("x=1.0"),
    )

    assert tuple(mb) == (mb.opening, *mb.body, mb.closing)
    assert mb.has_errors


def test_not_proper_statement():
    class MySt(fp.ParsedStatement):
        pass

    with pytest.raises(NotImplementedError):
        MySt.from_string("a = 1")

    with pytest.raises(NotImplementedError):
        MySt.from_string_and_config("a = 1", None)



from __future__ import annotations

import hashlib
import typing

import pytest

import flexparser.flexparser as fp
from flexparser.flexparser import _HASH_ALGORITHMS


def test_yield_types():
    class X:
        pass

    assert tuple(fp._yield_types(float)) == (float,)
    assert tuple(fp._yield_types(X)) == (X,)
    assert tuple(fp._yield_types(X())) == ()


def test_yield_types_container():
    class X:
        pass

    o = tuple[float, X]
    assert tuple(fp._yield_types(o)) == (float, X)

    o = tuple[float, ...]
    assert tuple(fp._yield_types(o)) == (float,)

    o = tuple[typing.Union[float, X], ...]
    assert tuple(fp._yield_types(o)) == (float, X)


def test_yield_types_union():
    class X:
        pass

    o = typing.Union[float, X]
    assert tuple(fp._yield_types(o)) == (float, X)


def test_yield_types_list():
    o = list[float]
    assert tuple(fp._yield_types(o)) == (float,)


def test_hash_object():
    content = b"spam \n ham"
    hasher = hashlib.sha1

    ho = fp.Hash.from_bytes(hashlib.sha1, content)
    hd = hasher(content).hexdigest()
    assert ho.algorithm_name == "sha1"
    assert ho.hexdigest == hd
    assert ho != hd
    assert ho != fp.Hash.from_bytes(hashlib.md5, content)
    assert ho == fp.Hash.from_bytes(hashlib.sha1, content)


@pytest.mark.parametrize("algo_name", _HASH_ALGORITHMS)
def test_hash_items(algo_name: str):
    content = b"spam \n ham"
    hasher = getattr(hashlib, algo_name)

    ho = fp.Hash.from_bytes(hasher, content)
    hd = hasher(content).hexdigest()
    assert ho.algorithm_name == algo_name
    assert ho.hexdigest == hd
    assert ho != hd
    assert ho == fp.Hash.from_bytes(hasher, content)
